;;;;SMEVAL.SCM

;;;                SUBSTITUTION MODEL for 6.001 Spring 95

;; This version handles SET! but not SET-CAR!  Yet it doesn't use
;; SET! (or data mutation).  In fact the implementation is purely
;; functional except for printing and generation of "fresh" symbols.

;;; This file defines the EVAL-PRINT loop of the Substitution Model .

;; The procedure SMSTEP-STREAM defines the stream of successively 
;; rewritten bodies generated by successive calls of the procedure
;; ONE-STEP-BODY on a body given in kernel syntax.  The stream
;; continues until ONE-STEP-BODY reaches a value or gets stuck,
;; which is indicated by a final tag.

;; The main procedure of the Scheme Term Rewriting Model implementation
;; is ONE-STEP-BODY defined in the file SMSTEP.SCM.  ONE-STEP-BODY does
;; a single rewrite step, if possible, on a body.  It satisfies the
;; following "contract":

;;;                         CONTRACT FOR ONE-STEP-BODY

;;In a call (ONE-STEP-BODY body defines-from-context):

;;           BODY must be desugared.
;;           If a rewrite rule applies to BODY, using DEFINES-FROM-CONTEXT for
;;           lookup of undefined variables, then the call returns the tagged result
;;                  ('stepped, rewritten body) unless
;;           the rewrite rule was a SET! whose variable is not defined at the
;;           beginning of BODY, in which case it returns
;;                  ('set!, (SET!-INFO, rewritten body))
;;           where SET!-INFO is (variable-to-be-set, <immediate value> to set it to).
;;           If body is a syntactic value, then it returns
;;                  ('val, garbage collected body).
;;           If body cannot be rewritten and is not a value, then it returns
;;                  ('stuck, info on the problem).

;;;                     TYPE INFORMATION FOR ONE-STEP-BODY

;; ONE-STEP-BODY: (<body>,List(<define>)) --> Tagged-Body
;; Tagged-Body  =  Labelled-Sum[stepped: <body>,
;;                                 set!: Set!-info X <body>,
;;                                  val: <body>,
;;                                stuck: Error-info]
;; Set!-info  =  Var X <immediate value>


;;SMSTEP-STREAM: <body> --> (Stream(<body>)  X_str {val, stuck})

(define (smstep-stream body)
  (let ((tagged-body (one-step-body body global-definitions)))
    (let ((tagged-body (handle-toplevel-set! tagged-body)))
      (let ((tag (tag-of tagged-body))
            (next-body-or-info (datum-of-tagged-body tagged-body)))
        (if (eq? tag 'stuck)
            (stream next-body-or-info 'stuck)
            (let ((next-body (maybe-collect-garbage next-body-or-info)))
              (cons-stream
               next-body
               (if (eq? tag 'val)
                   (stream 'val)
                   (smstep-stream next-body)))))))))


(define (handle-toplevel-set! tagged-body)
  (let ((tag (tag-of tagged-body)))
    (if (eq? tag 'set!)
        (let ((datum (datum-of-tagged-body tagged-body)))
          (let ((var (caar datum)))
            (let ((global-def (lookup-define var global-definitions)))
              (if global-def            ;shadow global def with new def from SET!
                  (let ((body (cdr datum)))
                    (make-tagged-body
                     'stepped
                     (make-body
                      (cons
                       (make-define
                        var
                        (cdar datum))
                       (defines-of-body body))
                      (expression-of-body body))))                
                  (make-tagged-body 'stuck (list 'unbound-variable: var))))))
        tagged-body)))


(define (maybe-collect-garbage body)
  (if (garbage-collect-this-step? body)
      (garbage-collect
       (defines-of-body body)
       (expression-of-body body))
      body))


    ;;GARBAGE COLLECTION IMPOSED-RATE PREDICATE

;always impose a garbage collection
(define (garbage-collect-this-step? body) #t)

;don't impose extra garbage collections
;(define (garbage-collect-this-step? body) #f)

;impose a garbage collection every 50th step
      ;;A SET! is used here for convenience. We could
      ;;have eliminated it by adding a STEP-NUMBER parameter
      ;;to SMSTEP-STREAM.  (This is the only place a SET!
      ;;form is used in the entire implementation of the 
      ;;Substitution Model.)
;(define garbage-collect-this-step?
;  (let ((count 0))
;    (lambda (body)
;      (set! count (modulo (inc count) 50))
;      (zero? count))))


;;;THE GLOBAL DEFINITIONS: List(<immediate defines>)
;;;These are all user-definable, but are supplied for convenience
;;;and to minimize clutter in the printout.  By being installed here,
;;;they aren't printed.

;;;They must be desugared before being included in the list.

(define global-definitions
  '((define (member e l)
      (if (null? l)
          #f
          (if (equal? e (car l))
              l
              (member e (cdr l)))))
    (define (memq e l)
      (if (null? l)
          #f
          (if (eq? e (car l))
              l
              (memq e (cdr l)))))
    (define (for-all? l pred)
      (if (null? l)
          #t
          ((lambda (x rest)
             (if (null? rest)
                 x
                 (if x
                     (for-all? rest pred)
                     #f)))
           (pred (car l))
           (cdr l))))))


                               ;;;;CONTROLLING THE PRINTOUT

;;SMEVAL: <body> --> final-message
;;nicely prints out (SMSTEP-STREAM BODY)


(define (smeval body)
  (define (print-steps step-number body steps)
    (if (interrupt-smeval? step-number body)
        (final-message 'interrupted)
        (let ((next-body (stream-car steps))
              (rest-steps (stream-cdr steps)))
          (if (stream-null? (stream-cdr rest-steps))
              (let ((tag (stream-car rest-steps)))
                (cond ((eq? tag 'val)
                                        ;NEXT-BODY and BODY only differ by
                                        ;garbage collection
                       (if (or (not (equal? next-body body))
                               (not (print-this-step? step-number body)))
                           (print-stepped-message step-number next-body))
                       (maybe-print-printable-version next-body))
                      (else (print-stepped-message step-number next-body)))
                (final-message tag))
              (begin
                (if (print-this-step? step-number next-body)
                    (print-stepped-message step-number next-body))
                (print-steps (inc step-number) next-body rest-steps))))))
  (let ((desugared-body (desugar body)))
    (if (not (equal? body desugared-body))
        (print-stepped-message 0 desugared-body))
    (print-steps 1 desugared-body (smstep-stream desugared-body))))
                   

    ;;SUBMODEL INTERRUPT PREDICATE

;(define (interrupt-smeval? step-number body) #f)  ;don't interrupt

(define (interrupt-smeval? step-number body)        ;interrupt after 600 steps
  (> step-number 600))


     ;;PRINT STEP PREDICATE

;;see the file PRNTLESS.SCM for other printing rates
(define (print-this-step? n body) #t)

(define (print-stepped-message step-number body)
  (begin (newline)
         (newline)
         (display ";==(")
         (display step-number)
         (display ")==>")
         (pp body)))

(define (maybe-print-printable-version body)
  (let ((output (printable-version body)))
    (if (not (equal? body output))
        (begin (newline) (pp output)))))

(define (final-message tag)
   (cond
    ((eq? tag 'val) "Syntactic Value was returned")
    ((eq? tag 'interrupted) "Rewriting got interrupted")
    ((eq? tag 'stuck) "Rewriting got stuck")))


     ;; CONVERT SUB-MODEL VALUE to SCHEME OUTPUT

    ;;PRINTABLE-VERSION: <value> --> Nested-List(Primitive-value)
;Primitive-value = Sch-Symbol U Sch-Num U Sch-Bool U {'()} U String
;Nexted-List(T)  =  T  U  List(Nested-List(T))

(define (printable-version value)
  (cond ((and (combination? value)
              (eq? 'cons (operator-of value)))
         (let ((rands (operands-of value)))
           (if (submodel-null? (cadr rands))
               (list (printable-version (car rands)))
               (cons
                (printable-version (car rands))
                (printable-version (cadr rands))))))
        ((symbol-expression? value)
         (symbol-name value))
        ((submodel-null? value)
         '())
        ((or (lambda-expression? value)
             (primitive-procedure-variable? value)
             (rule-specified-procedure-variable? value))
         (cons 'procedure-object: value))
        ((body-with-defines? value)
         (printable-version (expression-of-body value))) ;;don't print the defines
        (else value)))                  ;self-evaluating or unknown, so leave it alone


;
